function [rmd, gamma] = rot_euc(p, q, C, lambda, options)
%ROT_EUC    Regularized Optimal Transport with Euclidean Distance.
%   [rmd, gamma] = rot_euc(p, q, C, lambda, options) computes the  
%   regularized optimal transport based on the Euclidean distance generated
%   by a quadratic cost function. Dykstra's algorithm with alternate 
%   Bregman projections is used. This reduces to an analytical algorithm.
%
%   Input arguments:
%       * p: probability vector from which mass is transported.
%       * q: probability vector to which mass is transported.
%       * C: non-negative cost matrix.
%       * lambda: positive scalar penalty.
%       * options: structure of options with the following fields:
%           - verbose: boolean flag to display information on progress (default: false);
%           - check_argin: boolean flag to check input arguments or not (default: true);
%           - max_iter: maximum number of iterations for the loop of Dykstra's algorithm (default: 100);
%           - term_iter: number of iterations after which termination is checked (default: 1);
%           - term_tol: positive tolerance of termination in marginal difference (default: 1e-4);
%           - term_crit: criterion of termination between 'rmd' in relative variation, 'gamma' in absolute variation, and 'pq' in marginal difference (default: 'pq');
%           - term_norm: positive norm parameter for the criterion of termination (default: inf).
%
%   Output arguments:
%       * rmd: rot mover's distance;
%       * gamma: optimal transport plan.
%
%   Copyright (c) 2016, Arnaud Dessein (Universit de Bordeaux)

%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Process input arguments %
%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Number of arguments
if nargin == 4
    options = struct;
elseif nargin < 4
    error('MATLAB:ROT:nargin', 'Not enough input arguments.');
end

% Flag to check input arguments
if isstruct(options)
    if isfield(options, 'check_argin')
        if ~(isscalar(options.check_argin) && islogical(options.check_argin))
            error('MATLAB:ROT:optionsCheckArgin', 'options.check_argin should be true or false.');
        end
    else
        options.check_argin = true;
    end
else
    error('MATLAB:ROT:options', 'options should be a structure.');
end

% Check input arguments
if options.check_argin

    % Probability vectors
    if ~(isvector(p) && isreal(p)) || any(p < 0) || abs(sum(p) - 1) > 1e-12
        error('MATLAB:ROT:p', 'p should be a probability vector.');
    end
    if ~(isvector(q) && isreal(q)) || any(q < 0) || abs(sum(q) - 1) > 1e-12
        error('MATLAB:ROT:q', 'q should be a probability vector.');
    end
    
    % Cost matrix
    if ~(isnumeric(C) && ndims(C) == 2 && size(C, 1) == length(p) && size(C, 2) == length(q) && isreal(C)) || any(C(:) < 0 | C(:) == inf)
        error('MATLAB:ROT:C', 'C should be a non-negative matrix with same dimensions as p and q.');
    end
    
    % Penalty
    if ~(isscalar(lambda) && isreal(lambda) && isfinite(lambda) && lambda > 0)
        error('MATLAB:ROT:lamdba', 'lambda should be a positive scalar.');
    end
    
    % Verbose
    if isfield(options, 'verbose')
        if ~(isscalar(options.verbose) && islogical(options.verbose))
            error('MATLAB:ROT:optionsVerbose', 'options.verbose should be true or false.');
        end
    else
        options.verbose = false;
    end
    
    % Maximum number of iterations
    if isfield(options, 'max_iter_main')
        if ~(isscalar(options.max_iter_main) && isreal(options.max_iter_main) && options.max_iter_main > 0 && round(options.max_iter_main) == options.max_iter_main)
            error('MATLAB:ROT:optionsMaxIter', 'options.max_iter should be a positive integer or infinite.');
        end
    else
        options.max_iter_main = 100;
    end
    
    % Number of iterations after which termination is checked
    if isfield(options, 'term_iter')
        if ~(isscalar(options.term_iter) && isreal(options.term_iter) && options.term_iter > 0 && round(options.term_iter) == options.term_iter && options.term_iter < options.max_iter)
            error('MATLAB:ROT:optionsTermIter', 'options.term_iter should be a positive integer less than options.max_iter.');
        end
    else
        options.term_iter = 1;
    end
    
    % Termination tolerance
    if isfield(options, 'term_tol')
        if ~(isscalar(options.term_tol) && isreal(options.term_tol) && isfinite(options.term_tol) && options.term_tol > 0)
            error('MATLAB:ROT:optionsTermTol', 'options.term_tol should be a positive scalar.');
        end
    else
        options.term_tol = 1e-4;
    end
    
    % Norm parameter for termination criterion
    if isfield(options, 'term_norm')
        if ~(isscalar(options.term_norm) && isreal(options.term_norm) && options.term_norm > 0)
            error('MATLAB:ROT:optionsTermNorm', 'options.term_norm should be a positive scalar or infinite.');
        end
    else
        options.term_norm = inf;
    end
    
end

% Unit entries
punit = find(p == 1);
qunit = find(q == 1);
if ~isempty(punit)
    qt = q.';
    if nargout == 2
        gamma           = zeros(length(p), length(q));
        gamma(punit, :) = qt;
    end
    rmd = sum(qt .* C(punit, :));
    return
end
if ~isempty(qunit)
    if nargout == 2
        gamma           = zeros(length(p), length(q));
        gamma(:, qunit) = p;
    end
    rmd = sum(p .* C(:, qunit));
    return
end

% Null entries
if nargout == 2
    gamma = zeros(length(p), length(q));
end
ppos    = find(p ~= 0);
qpos    = find(q ~= 0);
p       = p(ppos);
q       = q(qpos);
C       = C(ppos, qpos);
m       = length(p);
n       = length(q);
qt      = q.';

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Regularized optimal transport with Kullback-Leibler divergence %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Initialization
xi          = C / -lambda;
psip        = zeros(m, n);
term_flag   = false;

if strcmp(options.term_crit, 'rmd')
    rmd_old_main    = sum(psip(:) .* C(:));
elseif strcmp(options.term_crit, 'gamma')
    psip_old_main   = psip;
end

% Dykstra's algorithm for alternate Bregman projections with Euclidean distance
for k = 1: options.max_iter_main

    % Verbose
    if options.verbose
        fprintf('Iteration %d.\n', k);
    end
        
    % Row offsetting
    xi = bsxfun(@plus, xi, (p - sum(psip, 2) - p) / n);
        
    % Entry truncation
    psip = max(0, xi);
    
    % Column offsetting
    xi = bsxfun(@plus, xi, (qt - sum(psip, 1)) / m);
    
    % Entry truncation
    psip = max(0, xi);
    
    % Check termination
    if mod(k, options.term_iter_main) == 0 && k ~= options.max_iter_main      
        if strcmp(options.term_crit, 'rmd')
            rmd     = sum(psip(:) .* C(:));
            crit    = abs(rmd / rmd_old_main - 1);
            if options.verbose
                fprintf('Relative variation in rmd: %e.\n', crit);
            end
            if crit < options.term_tol_main || (rmd_old_main == 0 && rmd == 0)
                term_flag_main = true;
                break
            end
            rmd_old_main = rmd;
        elseif strcmp(options.term_crit, 'gamma')
            crit = norm(((psip_old_main(:)-psip(:))), options.term_norm)./norm(psip_old_main(:), options.term_norm);
            if options.verbose
                fprintf('Absolute variation in gamma: %e.\n', crit);
            end
            if crit < options.term_tol_main
                term_flag_main = true;
                break
            end
            psip_old_main = psip;
        else
            crit = norm([sum(psip, 2); sum(psip, 1).'] - [p; q], options.term_norm);
            
            if options.verbose
                fprintf('Marginal difference in p and q: %e.\n', crit);
            end
            if crit < options.term_tol_main
                term_flag_main = true;
                break
            end
            
        end
    end
    
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Format output arguments %
%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Rot mover's distance
rmd = sum(psip(:) .* C(:));
if options.verbose
    if term_flag
        fprintf('Iterations terminated.\n');
    else
        fprintf('Iterations finished.\n');
    end
end

% Optimal transport plan
if nargout == 2
    gamma(ppos, qpos) = psip;
end