function [rmd, gamma] = rot_hell(p, q, C, lambda, options)
%ROT_HELL	Regularized Optimal Transport with Hellinger Distance.
%   [rmd, gamma] = rot_hell(p, q, C, lambda, options) computes the
%   regularized optimal transport based on the Hellinger distance.
%   Dykstra's algorithm with alternate Bregman projections is used. The
%   Newton-Raphson method is used to find the projections.
%
%   Input arguments:
%       * p: probability vector from which mass is transported.
%       * q: probability vector to which mass is transported.
%       * C: non-negative cost matrix.
%       * lambda: positive scalar penalty.
%       * options: structure of options with the following fields:
%           - verbose: boolean flag to display information on progress (default: false);
%           - check_argin: boolean flag to check input arguments or not (default: true);
%           - max_iter_main: maximum number of iterations for the main outer loop of Dykstra's algorithm (default: 100);
%           - max_iter_aux: maximum number of iterations for the auxiliary inner loops of the Newton-Raphson method (default: 10);
%           - term_iter_main: number of main iterations after which termination is checked (default: 1);
%           - term_iter_aux: number of auxiliary iterations after which termination is checked (default: 1);
%           - term_tol_main: main tolerance of termination in marginal difference (default: 1e-4);
%           - term_tol_aux: auxiliary tolerance of termination in marginal difference (default: 1e-4);
%           - term_crit: criterion of termination between 'rmd' in relative variation, 'gamma' in absolute variation, and 'pq' in marginal difference (default: 'pq');
%           - term_norm: positive norm parameter for the criterion of termination (default: inf).
%
%   Output arguments:
%       * rmd: rot mover's distance;
%       * gamma: optimal transport plan.
%
%   Copyright (c) 2016, Arnaud Dessein (Universit de Bordeaux)

%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Process input arguments %
%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Number of arguments
if nargin == 4
    options = struct;
elseif nargin < 4
    error('MATLAB:ROT:nargin', 'Not enough input arguments.');
end

% Flag to check input arguments
if isstruct(options)
    if isfield(options, 'check_argin')
        if ~(isscalar(options.check_argin) && islogical(options.check_argin))
            error('MATLAB:ROT:optionsCheckArgin', 'options.check_argin should be true or false.');
        end
    else
        options.check_argin = true;
    end
else
    error('MATLAB:ROT:options', 'options should be a structure.');
end

% Check input arguments
if options.check_argin
    
    % Probability vectors
    if ~(isvector(p) && isreal(p)) || any(p < 0) || abs(sum(p) - 1) > 1e-12
        error('MATLAB:ROT:p', 'p should be a probability vector.');
    end
    if ~(isvector(q) && isreal(q)) || any(q < 0) || abs(sum(q) - 1) > 1e-12
        error('MATLAB:ROT:q', 'q should be a probability vector.');
    end
    
    % Dimensions
    m = length(p);
    n = length(q);
    
    % Cost matrix
    if ~(isnumeric(C) && ndims(C) == 2 && size(C, 1) == m && size(C, 2) == n && isreal(C)) || any(C(:) < 0 | C(:) == inf)
        error('MATLAB:ROT:C', 'C should be a non-negative matrix with same dimensions as p and q.');
    end
    
    % Penalty
    if ~(isscalar(lambda) && isreal(lambda) && isfinite(lambda) && lambda > 0)
        error('MATLAB:ROT:lamdba', 'lambda should be a positive scalar.');
    end
    
    % Verbose
    if isfield(options, 'verbose')
        if ~(isscalar(options.verbose) && islogical(options.verbose))
            error('MATLAB:ROT:optionsVerbose', 'options.verbose should be true or false.');
        end
    else
        options.verbose = false;
    end
    
    % Maximum number of main iterations
    if isfield(options, 'max_iter_main')
        if ~(isscalar(options.max_iter_main) && isreal(options.max_iter_main) && options.max_iter_main > 0 && round(options.max_iter_main) == options.max_iter_main)
            error('MATLAB:ROT:optionsMaxIterMain', 'options.max_iter_main should be a positive integer or infinite.');
        end
    else
        options.max_iter_main = 100;
    end
    
    % Maximum number of auxiliary iterations
    if isfield(options, 'max_iter_aux')
        if ~(isscalar(options.max_iter_aux) && isreal(options.max_iter_aux) && options.max_iter_aux > 0 && round(options.max_iter_aux) == options.max_iter_aux)
            error('MATLAB:ROT:optionsMaxIterAux', 'options.max_iter_aux should be a positive integer or infinite.');
        end
    else
        options.max_iter_aux = 10;
    end
    
    % Number of main iterations after which termination is checked
    if isfield(options, 'term_iter_main')
        if ~(isscalar(options.term_iter_main) && isreal(options.term_iter_main) && options.term_iter_main > 0 && round(options.term_iter_main) == options.term_iter_main && options.term_iter_main < options.max_iter_main)
            error('MATLAB:ROT:optionsTermIterMain', 'options.term_iter_main should be a positive integer less than options.max_iter_main.');
        end
    else
        options.term_iter_main = 1;
    end
    
    % Number of auxiliary iterations after which termination is checked
    if isfield(options, 'term_iter_aux')
        if ~(isscalar(options.term_iter_aux) && isreal(options.term_iter_aux) && options.term_iter_aux > 0 && round(options.term_iter_aux) == options.term_iter_aux && options.term_iter_aux < options.max_iter_aux)
            error('MATLAB:ROT:optionsTermIterAux', 'options.term_iter_aux should be a positive integer less than options.max_iter_aux.');
        end
    else
        options.term_iter_aux = 1;
    end
    
    % Main termination tolerance
    if isfield(options, 'term_tol_main')
        if ~(isscalar(options.term_tol_main) && isreal(options.term_tol_main) && isfinite(options.term_tol_main) && options.term_tol_main > 0)
            error('MATLAB:ROT:optionsTermTolMain', 'options.term_tol_main should be a positive scalar.');
        end
    else
        options.term_tol_main = 1e-4;
    end
    
    % Auxiliary termination tolerance
    if isfield(options, 'term_tol_aux')
        if ~(isscalar(options.term_tol_aux) && isreal(options.term_tol_aux) && isfinite(options.term_tol_aux) && options.term_tol_aux > 0)
            error('MATLAB:ROT:optionsTermTolAux', 'options.term_tol_aux should be a positive scalar.');
        end
    else
        options.term_tol_aux = 1e-4;
    end
    
   % Termination criterion
    if isfield(options, 'term_crit')
        if ~(strcmp(options.term_crit, 'rmd') || strcmp(options.term_crit, 'gamma') || strcmp(options.term_crit, 'pq'))
            error('MATLAB:ROT:optionsTermCrit', 'options.term_crit should be ''rmd'', ''gamma'', or ''pq''.');
        end
    else
        options.term_crit = 'pq';
    end  
    
    % Norm parameter for termination criterion
    if isfield(options, 'term_norm')
        if ~(isscalar(options.term_norm) && isreal(options.term_norm) && options.term_norm > 0)
            error('MATLAB:ROT:optionsTermNorm', 'options.term_norm should be a positive scalar or infinite.');
        end
    else
        options.term_norm = inf;
    end
    
end

% Unit entries
punit = find(p == 1);
qunit = find(q == 1);
if ~isempty(punit)
    qt = q.';
    if nargout == 2
        gamma           = zeros(length(p), length(q));
        gamma(punit, :) = qt;
    end
    rmd = sum(qt .* C(punit, :));
    return
end
if ~isempty(qunit)
    if nargout == 2
        gamma           = zeros(length(p), length(q));
        gamma(:, qunit) = p;
    end
    rmd = sum(p .* C(:, qunit));
    return
end

% Null entries
if nargout == 2
    gamma = zeros(length(p), length(q));
end
ppos    = find(p ~= 0);
qpos    = find(q ~= 0);
p       = p(ppos);
q       = q(qpos);
C       = C(ppos, qpos);
m       = length(p);
n       = length(q);
qt      = q.';

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Regularized optimal transport with Hellinger distance %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Initialization
theta           = C / -lambda;
thetap          = max(0, theta);
pon             = p / n;
qtom            = qt / m;
ponphip         = pon ./ sqrt(1 - pon.^2);
qtomphip        = qtom ./ sqrt(1 - qtom.^2);
term_flag_main  = false;
term_flag_aux   = false;
psip = thetap ./ (1 + thetap.^2);
if strcmp(options.term_crit, 'rmd')
    rmd_old_main    = sum(psip(:) .* C(:));
elseif strcmp(options.term_crit, 'gamma')
    psip_old_main   = psip;
end
% Dykstra's algorithm for alternate Bregman projections with Hellinger distance
for k = 1: options.max_iter_main
    
    % Verbose
    if options.verbose
        fprintf('Iteration %d.\n', k);
    end
    
    % Binary search for first Bregman projection with Hellinger distance
    taus            = sort(thetap, 2, 'ascend');
    taus            = cat(2, taus(:, 1) - ponphip, taus);
    rows            = (1: m).';
    mincols         = ones(m, 1);
    maxcols         = repmat(n, m, 1);
    while any(maxcols - mincols > 1)
        cols            = floor((mincols + maxcols) / 2);
        tautest         = taus(rows + m * (cols - 1));
        thetatest       = bsxfun(@plus, thetap, -tautest);
        psiptest        = thetatest ./ (1 + thetatest.^2);
        test            = sum(psiptest, 2) - p > 0;
        mincols(test)   = cols(test);
        maxcols(~test)  = cols(~test);
    end
    
    % Initialization
    tau         = (taus(rows + m * (mincols - 1)) + taus(rows + m * (maxcols - 1))) / 2;
    thetatmp    = bsxfun(@plus, thetap, -tau);
    psipp       = 1 ./ (1 + thetatmp.^2);
    psip        = thetatmp .* psipp;
    
    % Newton-Raphson method for first Bregman projection with Hellinger distance
    for l = 1: options.max_iter_aux
        
        % Verbose
        if options.verbose
            fprintf('\tIteration %d.\n', l);
        end
        
        % Update
        psipp       = psipp .* sqrt(psipp);
        tau         = tau + (sum(psip, 2) - p) ./ sum(psipp, 2);
        thetatmp    = bsxfun(@plus, thetap, -tau);
        psipp       = 1 ./ (1 + thetatmp.^2);
        psip        = thetatmp .* psipp;
        
        % Check termination
        if mod(l, options.term_iter_aux) == 0 && l ~= options.max_iter_aux
            crit = norm(sum(psip, 2) - p, options.term_norm);
            if options.verbose
                fprintf('\tMarginal difference in p: %e.\n', crit);
            end
            if crit < options.term_tol_aux
                term_flag_aux = true;
                break
            end
        end
        
    end
    
    % Update
    if term_flag_aux
        term_flag_aux = false;
        if options.verbose
            fprintf('\tIterations terminated.\n');
        end
    else
        if options.verbose
            fprintf('\tIterations finished.\n');
        end
    end
    
    % Entry truncation for non-negative projection
    theta   = bsxfun(@plus, theta, -tau);
    thetap  = max(0, theta);
    
    % Binary search for second Bregman projection with Hellinger distance
    sigmast         = sort(thetap, 1, 'ascend');
    sigmast         = cat(1, sigmast(1, :) - qtomphip, sigmast);
    cols            = 1: n;
    minrows         = ones(1, n);
    maxrows         = repmat(m, 1, n);
    while any(maxrows - minrows > 1)
        rows            = floor((minrows + maxrows) / 2);
        sigmattest      = sigmast(rows + m * (cols - 1));
        thetatest       = bsxfun(@plus, thetap, -sigmattest);
        psiptest        = thetatest ./ (1 + thetatest.^2);
        test            = sum(psiptest, 1) - qt > 0;
        minrows(test)   = rows(test);
        maxrows(~test)  = rows(~test);
    end
    
    % Initialization
    sigmat      = (sigmast(minrows + m * (cols - 1)) + sigmast(maxrows + m * (cols - 1))) / 2;
    thetatmp    = bsxfun(@plus, thetap, -sigmat);
    psipp       = 1 ./ (1 + thetatmp.^2);
    psip        = thetatmp .* psipp;
    
    % Newton-Raphson method for second Bregman projection with Hellinger distance
    for l = 1: options.max_iter_aux
        
        % Verbose
        if options.verbose
            fprintf('\tIteration %d.\n', l);
        end
        
        % Update
        psipp       = psipp .* sqrt(psipp);
        sigmat      = sigmat + (sum(psip, 1) - qt) ./ sum(psipp, 1);
        thetatmp    = bsxfun(@plus, thetap, -sigmat);
        psipp       = 1 ./ (1 + thetatmp.^2);
        psip        = thetatmp .* psipp;
        
        % Check termination
        if mod(l, options.term_iter_aux) == 0 && l ~= options.max_iter_aux
            crit = norm(sum(psip, 1) - qt, options.term_norm);
            if options.verbose
                fprintf('\tMarginal difference in q: %e.\n', crit);
            end
            if crit < options.term_tol_aux
                term_flag_aux = true;
                break
            end
        end
        
    end
    
    % Update
    if term_flag_aux
        term_flag_aux = false;
        if options.verbose
            fprintf('\tIterations terminated.\n');
        end
    else
        if options.verbose
            fprintf('\tIterations finished.\n');
        end
    end
    
    % Entry truncation for non-negative projection
    theta   = bsxfun(@plus, theta, -sigmat);
    thetap  = max(0, theta);
    
    % Check termination
    if mod(k, options.term_iter_main) == 0 && k ~= options.max_iter_main
        psip = thetap ./ (1 + thetap.^2);
        
        if strcmp(options.term_crit, 'rmd')
            rmd     = sum(psip(:) .* C(:));
            crit    = abs(rmd / rmd_old_main - 1);
            if options.verbose
                fprintf('Relative variation in rmd: %e.\n', crit);
            end
            if crit < options.term_tol_main || (rmd_old_main == 0 && rmd == 0)
                term_flag_main = true;
                break
            end
            rmd_old_main = rmd;
        elseif strcmp(options.term_crit, 'gamma')
            crit = norm(((psip_old_main(:)-psip(:))), options.term_norm)./norm(psip_old_main(:), options.term_norm);
            if options.verbose
                fprintf('Absolute variation in gamma: %e.\n', crit);
            end
            if crit < options.term_tol_main
                term_flag_main = true;
                break
            end
            psip_old_main = psip;
        else
            crit = norm([sum(psip, 2); sum(psip, 1).'] - [p; q], options.term_norm);
            
            if options.verbose
                fprintf('Marginal difference in p and q: %e.\n', crit);
            end
            if crit < options.term_tol_main
                term_flag_main = true;
                break
            end
            
        end
    end
    
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Format output arguments %
%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Rot mover's distance
if term_flag_main
    if options.verbose
        fprintf('Iterations terminated.\n');
    end
else
    psip = thetap ./ (1 + thetap.^2);
    if options.verbose
        fprintf('Iterations finished.\n');
    end
end
rmd = sum(psip(:) .* C(:));

% Optimal transport plan
if nargout == 2
    gamma(ppos, qpos) = psip;
end