%% Reset

%Examples of ROT mover's distances.
%If using, please cite


%Dessein, A., Papadakis, N., & Rouas, J. L. (2018). Regularized optimal transport and the rot mover's distance. The Journal of Machine Learning Research, 19(1), 590-642.


clear all
close all
clc

addpath('./src')


%% Synthetic data

% Problem dimension
d = 128;

% Probability vectors
x = (0: 1 / (d - 1): 1).';
p = exp(-(x - 0.5).^2 / 0.2^2);
p = p / sum(p(:));
q = exp(-(x - 0.25).^2 / 0.1^2) + exp(-(x - 0.75).^2 / 0.1^2);
q = q / sum(q(:));

% Cost matrix
C = toeplitz(x.^2);
C(C == 0)   = 1e-12;
%% Parameters
% Options
options.verbose         = false;
options.check_argin     = true;
options.max_iter_main   = 1e10;
options.max_iter_aux    = 1e2;
options.term_iter_main  = 1;
options.term_iter_aux   = 1;

% The following parameters can be tunned for each ROT to have far better performances. 
% Here it gives fair results for all ROT

options.term_tol_main   = 1e-3;
options.term_tol_aux    = 5e-5;
options.term_crit       = 'gamma';
options.term_norm       = Inf;

%% Regularized optimal transport

disp('Optimal transport.')
tic
[emd, gamma] = ot(p, q, C);
toc

disp('Regularized optimal transport with Hellinger divergence.')
lambda = 5e-1;
tic
[rmd_hell, gamma_hell] = rot_hell(p, q, C, lambda, options);
toc

%%
disp('Regularized optimal transport with Itakura-Saito divergence.')
tic
lambda = 1e-6;
[rmd_is, gamma_is] = rot_is(p, q, C, lambda, options);
toc

disp('Regularized optimal transport with Euclidean distance.')
tic
lambda = 1e-1;
[rmd_euc, gamma_euc] = rot_euc(p, q, C, lambda, options);
toc

%%
disp('Regularized optimal transport with lpn divergence.')
tic
lambda = 1e-1;
[rmd_lpn, gamma_lpn] = rot_lpn(p, q, C, lambda, 2.,options);
toc

disp('Regularized optimal transport with lpqn divergence.')
tic
lambda = 1e-4;
[rmd_lpqn, gamma_lpqn] = rot_lpqn(p, q, C, lambda,0.5, options);
toc
%%
lambda = 5e-5;
disp('Regularized optimal transport with beta divergence.')
tic
[rmd_beta, gamma_beta] = rot_beta(p, q, C, lambda,0.5, options);
toc

lambda = 5e-4;
disp('Regularized optimal transport with KL divergence.')
tic
[rmd_kl, gamma_kl] = rot_kl(p, q, C, lambda, options);
toc

lambda = 5e-4;
disp('Regularized optimal transport with log divergence.')
tic
[rmd_log, gamma_log] = rot_log(p, q, C, lambda, options);
toc

%% Display

figure
hold on
plot(x, p, '-b')
plot(x, q, '-r')
legend('p', 'q');
title('Probability distributions')

figure
imagesc(C)
title('Cost matrix')
colorbar

figure
imagesc(p * q.')
title('pq''')
colorbar



figure
subplot 331
imagesc(gamma)
title('OT')

subplot 332
imagesc(gamma_hell)
title('ROT with Hellinger divergence')

subplot 333
imagesc(gamma_is)
title('ROT with Itakura-Saito divergence')

subplot 334
imagesc(gamma_euc)
title('ROT  with Euclidean distance')

subplot 335
imagesc(gamma_lpn)
title('ROT with lpn divergence')

subplot 336
imagesc(gamma_lpqn)
title('ROT with lpqn divergence')

subplot 337
imagesc(gamma_beta)
title('ROT with with beta divergence')

subplot 338
imagesc(gamma_kl)
title('ROT with KL divergence')

subplot 339
imagesc(gamma_log)
title('ROT with log divergence')



